import { Server } from 'socket.io'
import { TSocket, ExtendedCharacter } from '../../utilities/Types'
import ZoneManager from '../../managers/ZoneManager'
import prisma from '../../utilities/Prisma'
import { AStar } from '../../utilities/Player/AStar'
import Rotation from '../../utilities/Player/Rotation'

const moveTokens = new Map<number, symbol>()

export default function setupCharacterMove(socket: TSocket, io: Server) {
  socket.on('character:move', async ({ position_x, position_y }: { position_x: number; position_y: number }) => {
    const { character } = socket
    if (!character) return console.error('character:move error', 'Character not found')

    moveTokens.set(character.id, Symbol('moveToken'))
    const grid = await ZoneManager.getGrid(character.zoneId)
    if (!grid?.length) return console.error('character:move error', 'Grid not found or empty')

    const start = { x: Math.floor(character.position_x), y: Math.floor(character.position_y) }
    const end = { x: Math.floor(position_x), y: Math.floor(position_y) }

    if (isObstacle(end, grid)) return socket.emit('character:moveError', 'Destination is an obstacle')

    const path = AStar.findPath(start, end, grid)
    if (!path.length) return socket.emit('character:moveError', 'No valid path found')

    character.isMoving = true
    io.in(character.zoneId.toString()).emit('character:moved', character)
    moveAlongPath(socket, io, path, grid).catch(console.error)
  })
}

async function moveAlongPath(socket: TSocket, io: Server, path: Array<{ x: number; y: number }>, grid: number[][]) {
  const { character } = socket
  if (!character) return

  const moveToken = moveTokens.get(character.id)
  const stepDuration = 250
  const updateInterval = 50

  for (let i = 0; i < path.length - 1; i++) {
    const startTime = Date.now()
    const start = path[i]
    const end = path[i + 1]

    while (Date.now() - startTime < stepDuration) {
      if (moveTokens.get(character.id) !== moveToken) return

      const progress = (Date.now() - startTime) / stepDuration
      const current = interpolatePosition(start, end, progress)

      if (isObstacle(current, grid)) {
        // If obstacle encountered, stop at the last valid position
        await updateCharacter(character, start, Rotation.calculate(start.x, start.y, end.x, end.y))
        io.in(character.zoneId.toString()).emit('character:moved', character)
        return
      }

      await updateCharacter(character, current, Rotation.calculate(start.x, start.y, end.x, end.y))
      io.in(character.zoneId.toString()).emit('character:moved', character)
      await new Promise((resolve) => setTimeout(resolve, updateInterval))
    }
  }

  // Ensure the character reaches the exact final position
  if (moveTokens.get(character.id) === moveToken) {
    const finalPosition = path[path.length - 1]
    await updateCharacter(character, finalPosition, character.rotation)
    character.isMoving = false
    moveTokens.delete(character.id)
    io.in(character.zoneId.toString()).emit('character:moved', character)
  }
}

const interpolatePosition = (start: { x: number; y: number }, end: { x: number; y: number }, progress: number) => ({
  x: start.x + (end.x - start.x) * progress,
  y: start.y + (end.y - start.y) * progress
})

const isObstacle = ({ x, y }: { x: number; y: number }, grid: number[][]) => {
  const gridX = Math.floor(x)
  const gridY = Math.floor(y)
  return grid[gridY]?.[gridX] === 1 ||
    grid[gridY]?.[Math.ceil(x)] === 1 ||
    grid[Math.ceil(y)]?.[gridX] === 1 ||
    grid[Math.ceil(y)]?.[Math.ceil(x)] === 1
}

async function updateCharacter(character: ExtendedCharacter, { x, y }: { x: number; y: number }, rotation: number) {
  Object.assign(character, { position_x: x, position_y: y, rotation })
  ZoneManager.updateCharacterInZone(character.zoneId, character)
  await prisma.character.update({
    where: { id: character.id },
    data: { position_x: x, position_y: y, rotation }
  })
}
